import config

class Material():

    def __init__(self, from_config = True, E= None, Iy= None, Iz= None, G= None, J= None, A= None):

        if from_config:
            self.E = config.E
            self.Iy = config.Iy
            self.Iz = config.Iz
            self.G = config.G
            self.J = config.J
            self.A = config.A

        else:
            self.E = E
            self.Iy = Iy
            self.Iz = Iz
            self.G = G
            self.J = J
            self.A = A



import numpy as np

class Beam():

    def __init__(self, ID, node1, node2, material, A= None):

        self.ID = ID
        self.node1 = node1
        self.node2 = node2

        self.E = material.E
        self.Iy = material.Iy
        self.Iz = material.Iz
        self.G = material.G
        self.J = material.J

        if not A:
            self.A = material.A

        self.Ta = self.T()
    
    def L(self):
        return self.node1.distance(self.node2)
    
    def Kloc(self):
        """
        Local stiffness matrix
        """
        A = self.A
        E = self.E
        Iz = self.Iz
        Iy = self.Iy
        G = self.G
        J = self.J
        L = self.L()

        k = np.array([[A*E/L,  0,             0,             0,      0,            0,            -A*E/L, 0,             0,             0,      0,            0],
                    [0,      12*E*Iz/L**3,  0,             0,      0,            6*E*Iz/L**2,  0,      -12*E*Iz/L**3, 0,             0,      0,            6*E*Iz/L**2],
                    [0,      0,             12*E*Iy/L**3,  0,      -6*E*Iy/L**2, 0,            0,      0,             -12*E*Iy/L**3, 0,      -6*E*Iy/L**2, 0],
                    [0,      0,             0,             G*J/L,  0,            0,            0,      0,             0,             -G*J/L, 0,            0],
                    [0,      0,             -6*E*Iy/L**2,  0,      4*E*Iy/L,     0,            0,      0,             6*E*Iy/L**2,   0,      2*E*Iy/L,     0],
                    [0,      6*E*Iz/L**2,   0,             0,      0,            4*E*Iz/L,     0,      -6*E*Iz/L**2,  0,             0,      0,            2*E*Iz/L],
                    [-A*E/L, 0,             0,             0,      0,            0,            A*E/L,  0,             0,             0,      0,            0],
                    [0,      -12*E*Iz/L**3, 0,             0,      0,            -6*E*Iz/L**2, 0,      12*E*Iz/L**3,  0,             0,      0,            -6*E*Iz/L**2],
                    [0,      0,             -12*E*Iy/L**3, 0,      6*E*Iy/L**2,  0,            0,      0,             12*E*Iy/L**3,  0,      6*E*Iy/L**2,  0],
                    [0,      0,             0,             -G*J/L, 0,            0,            0,      0,             0,             G*J/L,  0,            0],
                    [0,      0,             -6*E*Iy/L**2,  0,      2*E*Iy/L,     0,            0,      0,             6*E*Iy/L**2,   0,      4*E*Iy/L,     0],
                    [0,      6*E*Iz/L**2,   0,             0,      0,            2*E*Iz/L,     0,      -6*E*Iz/L**2,  0,             0,      0,            4*E*Iz/L]])

        return k
    
 
    def T(self):
        """
        Transformation matrix
        """
        x1, y1, z1 = self.node1.coords
        x2, y2, z2 = self.node2.coords
        L = self.L()

        x = [(x2-x1)/L, (y2-y1)/L, (z2-z1)/L]
        if np.isclose(x1, x2) and np.isclose(z1, z2):
            if y2 > y1:
                y = [-1, 0, 0]
                z = [0, 0, 1]
            else:
                y = [1, 0, 0]
                z = [0, 0, 1]

        elif np.isclose(y1, y2):
            y = [0, 1, 0]
            z = np.cross(x, y)
            z = np.divide(z, np.sqrt(z[0]**2 + z[1]**2 + z[2]**2))

        else:
            proj = [x2-x1, 0, z2-z1]

            if y2 > y1:
                z = np.cross(proj, x)
            else:
                z = np.cross(x, proj)

            z = np.divide(z, np.sqrt(z[0]**2 + z[1]**2 + z[2]**2))
            y = np.cross(z, x)
            y = np.divide(y, np.sqrt(y[0]**2 + y[1]**2 + y[2]**2))

        dirCos = np.array([x, y, z])

        T = np.zeros((12, 12))
        T[0:3, 0:3] = dirCos
        T[3:6, 3:6] = dirCos
        T[6:9, 6:9] = dirCos
        T[9:12, 9:12] = dirCos

        return T
    
    def Kglob(self):
        """
        Global stiffness matrix
        """
        return np.matmul(np.matmul(np.linalg.inv(self.T()), self.Kloc()), self.T())

    
    def nodalLoadVector(self):
        """
        Nodal load vector in global coordinates
        """
        return np.matmul(np.matmul(self.Kloc(), np.linalg.inv(self.Ta)), self.D())
    
    def D(self):
        """
        Displacement vector in global coordinates
        """
        return np.append(self.node1.Displacements, self.node2.Displacements)
    
    def Dloc(self):
        """
        Displacement vector in local coordinates
        """
        return np.matmul(np.linalg.inv(self.Ta), self.D())
    
    def Floc(self):
        """
        Force vector in local coordinates
        """
        return np.matmul(self.Kloc(), self.Dloc())


import numpy as np
import networkx as nx
from scipy import sparse

from Spyffness.Beam import Beam
from Spyffness.Node import Node
from Spyffness.Material import Material


class Frame():

    def __init__(self):
        self.Nodes = {} # {ID: Node}
        self.Beams = {} # {ID: Beam}


    def addMaterial(self, form_config= True, E= None, Iy= None, Iz= None, G= None, J= None, A= None):
        if form_config:
            self.Material = Material(True)
        else:
            self.Material = Material(False, E, Iy, Iz, G, J, A)

    def addNode(self, ID, coords):
        if ID:
            self.Nodes[ID] = Node(ID, coords)
        else:
            self.Nodes[len(self.Nodes)] = Node(len(self.Nodes), coords)

    def addBeam(self, ID, node1, node2, A = None):
        if ID:
            self.Beams[ID] = Beam(ID, self.Nodes[node1], self.Nodes[node2], self.Material, A)
        else:
            self.Beams[len(self.Beams)] = Beam(len(self.Beams), self.Nodes[node1], self.Nodes[node2], self.Material, A)

    def newSections(self, A):
        for i, beam in enumerate(self.Beams):
            beam.A = A[i]
    
    def fromGraph(self, G, A = None):
        """
        The nodes are ordered on the z-axis. Member nodes are also ordered according to the z-coordinate.
        """
        sorted_nodes = sorted(G.nodes, key= lambda k: k[2])

        for i, node in enumerate(sorted_nodes):
            self.addNode(i, np.array(node))
        

        sorted_edges = [sorted([sorted(x, key= lambda k: k[2]) for x in G.edges()], key= lambda k: k[0][2])]
        for i, edge in enumerate(sorted_edges):
            node1= self.__getNodeIDfromCoords(edge[0])
            node2= self.__getNodeIDfromCoords(edge[1])
            self.addBeam(i, node1, node2, A)

    def __getNodeIDfromCoords(self, coords):
        for id, node in self.Nodes.items():
            if np.array_equal(node.coords, coords):
                return id 
            
    
    def K(self):
        n = len(self.Nodes)
        k = np.zeros((6*n, 6*n))
        for beam in self.Beams.values():
            kglob = beam.Kglob()
            i = beam.node1.ID
            j = beam.node2.ID
            k[6*i:6*i+6, 6*i:6*i+6] += kglob[:6,:6]
            k[6*j:6*j+6, 6*j:6*j+6] += kglob[6:,6:]
            k[6*i:6*i+6, 6*j:6*j+6] += kglob[:6,6:]
            k[6*j:6*j+6, 6*i:6*i+6] += kglob[6:,:6]

        return k
    
    def setLoads(self, node, load_vect):
        self.Nodes[node].Loads = load_vect
    
    def setSupport(self, node, support_vect):
        self.Nodes[node].support = support_vect
    
    def fixAllBottomNodes(self):
        bottom_coord = self.__findBottomNodesCoord()
        for node in self.Nodes.values():
            if node.coords[2] == bottom_coord:
                node.fixedSupport()

    def pinAllBottomNodes(self):
        for node in self.Nodes.values():
            if node.coords[2] == 0:
                node.pinnedSupport()

    def setCompresionLoad(self, load):
        top_coord = self.__findTopNodesCoord()
        for node in self.Nodes.values():
            if node.coords[2] == top_coord:
                node.Loads[2] = -load

    def __getSupports(self):
        supports = np.zeros(6*len(self.Nodes))
        for i, node in self.Nodes.items():
            supports[6*i:6*i+6] = node.support

        if np.count_nonzero(supports) == 0:
            raise ValueError("No supports have been set")
        
        return supports
    
    def __getLoads(self):
        loads = np.zeros(6*len(self.Nodes))
        for i, node in self.Nodes.items():
            loads[6*i:6*i+6] = node.Loads
        
        if np.count_nonzero(loads) == 0:
            raise ValueError("No loads have been set")
        
        return loads
    
    def __getDisplacements(self):
        displacements = np.zeros(6*len(self.Nodes))
        for i, node in self.Nodes.items():
            displacements[6*i:6*i+6] = node.Displacements

        return displacements
    
    def __getReactions(self):
        reactions = np.zeros(6*len(self.Nodes))
        for i, node in self.Nodes.items():
            reactions[6*i:6*i+6] = node.Reactions

        return reactions
    
    def solve(self):
        k = self.K()
        # print(sparse.csr_matrix(k))
        supports = self.__getSupports()
        loads = self.__getLoads()

        displacements = self.__getDisplacements()
        reactions = self.__getReactions()
        free = np.where(supports == 1)[0]
        fixed = np.where(supports == 0)[0]
        kfree = k[free,:][:,free]
        kfixed = k[fixed,:][:,fixed]

        loads = loads[free]
        displacements[free] = np.linalg.solve(kfree, loads)
        print(displacements)

        reactions[fixed] = np.dot(kfixed, displacements[fixed])
        print(reactions)
        self.__setDisplacements(displacements)
        self.__setReactions(reactions)

    def __setDisplacements(self, displacements):
        for i, node in self.Nodes.items():
            node.Displacements = displacements[6*i:6*i+6]

    
    def __setReactions(self, reactions):
        for i, node in self.Nodes.items():
            node.Reactions = reactions[6*i:6*i+6]

    def __findBottomNodesCoord(self):
        min_coord = 0
        for node in self.Nodes.values():
            if node.coords[2] < min_coord:
                min_coord = node.coords[2]

        return min_coord
    
    def __findTopNodesCoord(self):
        max_coord = 0
        for node in self.Nodes.values():
            if node.coords[2] > max_coord:
                max_coord = node.coords[2]

        return max_coord


import matplotlib.pyplot as plt


class Visualization():

    def __init__(self, model):
        self.Beams = model.Beams
        self.Nodes = model.Nodes


    def plot(self):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for beam in self.Beams.values():
            x = [beam.node1.coords[0], beam.node2.coords[0]]
            y = [beam.node1.coords[1], beam.node2.coords[1]]
            z = [beam.node1.coords[2], beam.node2.coords[2]]
            ax.plot(x, y, z)
        plt.show()

    def plotDisplacements(self, scale = 1):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for beam in self.Beams.values():
            x = [beam.node1.coords[0] + scale*beam.node1.Displacements[0], beam.node2.coords[0] + scale*beam.node2.Displacements[0]]
            y = [beam.node1.coords[1] + scale*beam.node1.Displacements[1], beam.node2.coords[1] + scale*beam.node2.Displacements[1]]
            z = [beam.node1.coords[2] + scale*beam.node1.Displacements[2], beam.node2.coords[2] + scale*beam.node2.Displacements[2]]
            ax.plot(x, y, z)
        plt.show()

    def plotReactions(self, scale = 1):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for beam in self.Beams.values():
            x = [beam.node1.coords[0] - scale*beam.node1.Reactions[0], beam.node2.coords[0] - scale*beam.node2.Reactions[0]]
            y = [beam.node1.coords[1] - scale*beam.node1.Reactions[1], beam.node2.coords[1] - scale*beam.node2.Reactions[1]]
            z = [beam.node1.coords[2] - scale*beam.node1.Reactions[2], beam.node2.coords[2] - scale*beam.node2.Reactions[2]]
            ax.plot(x, y, z)
        plt.show()

    def plotLoads(self, scale = 1):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for beam in self.Beams.values():
            x = [beam.node1.coords[0] + scale*beam.node1.Loads[0], beam.node2.coords[0] + scale*beam.node2.Loads[0]]
            y = [beam.node1.coords[1] + scale*beam.node1.Loads[1], beam.node2.coords[1] + scale*beam.node2.Loads[1]]
            z = [beam.node1.coords[2] + scale*beam.node1.Loads[2], beam.node2.coords[2] + scale*beam.node2.Loads[2]]
            ax.plot(x, y, z)
        plt.show()

    def plotSupports(self, scale = 1):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for beam in self.Beams.values():
            x = [beam.node1.coords[0] + scale*beam.node1.support[0], beam.node2.coords[0] + scale*beam.node2.support[0]]
            y = [beam.node1.coords[1] + scale*beam.node1.support[1], beam.node2.coords[1] + scale*beam.node2.support[1]]
            z = [beam.node1.coords[2] + scale*beam.node1.support[2], beam.node2.coords[2] + scale*beam.node2.support[2]]
            ax.plot(x, y, z)
        plt.show()

    


import numpy as np

class Node():


    def __init__(self, ID, coords):

        self.ID = ID
        self.coords = np.array(coords)

        self.Loads = np.zeros(6)
        self.Displacements = np.zeros(6)
        self.Reactions = np.zeros(6)

        self.support = np.ones(6)

    def distance(self, other):
        return np.linalg.norm(self.coords - other.coords)
    
    def fixedSupport(self):
        self.support = np.zeros(6)
    
    def pinnedSupport(self, thetax = 1, thetay = 1, thetaz = 1):
        self.support = np.array([0,0,0,thetax,thetay,thetaz])
    
